"""Definition for multiple victims that share a single GPU (sequentially)."""

import torch
import numpy as np

from collections import defaultdict
import copy

from .models import get_model
from ..hyperparameters import training_strategy
from ..utils import set_random_seed, average_dicts
from ..consts import BENCHMARK
from .context import GPUContext
torch.backends.cudnn.benchmark = BENCHMARK

from .victim_base import _VictimBase
from .training import get_optimizers

class _VictimEnsemble(_VictimBase):
    """Implement model-specific code and behavior for multiple models on a single GPU.

    --> Running in sequential mode!

    """

    """ Methods to initialize a model."""

    def initialize(self, pretrain=False, seed=None):
        if self.args.modelkey is None:
            if seed is None:
                self.model_init_seed = np.random.randint(0, 2**32 - 1)
            else:
                self.model_init_seed = seed
        else:
            self.model_init_seed = self.args.modelkey
        set_random_seed(self.model_init_seed)
        print(f'Initializing ensemble from random key {self.model_init_seed}.')

        self.models, self.definitions, self.optimizers, self.schedulers, self.epochs = [], [], [], [], []
        for idx in range(self.args.ensemble):
            model_name = self.args.net[idx % len(self.args.net)]
            model, defs, optimizer, scheduler = self._initialize_model(model_name, pretrain=pretrain)
            self.models.append(model)
            self.definitions.append(defs)
            self.optimizers.append(optimizer)
            self.schedulers.append(scheduler)
            print(f'{model_name} initialized as model {idx}')
            print(repr(defs))
        self.defs = self.definitions[0]

    def reinitialize_last_layer(self, reduce_lr_factor=1.0, seed=None, keep_last_layer=False):
        if self.args.modelkey is None:
            if seed is None:
                self.model_init_seed = np.random.randint(0, 2**32 - 1)
            else:
                self.model_init_seed = seed
        else:
            self.model_init_seed = self.args.modelkey
        set_random_seed(self.model_init_seed)


        for idx in range(self.args.ensemble):
            model_name = self.args.net[idx % len(self.args.net)]
            if not keep_last_layer:
                # We construct a full replacement model, so that the seed matches up with the initial seed,
                # even if all of the model except for the last layer will be immediately discarded.
                replacement_model = get_model(model_name, self.args.dataset, pretrained=self.args.pretrained_model)

                # Rebuild model with new last layer
                frozen = self.models[idx].frozen
                self.models[idx] = torch.nn.Sequential(*list(self.models[idx].children())[:-1], torch.nn.Flatten(),
                                                       list(replacement_model.children())[-1])
                self.models[idx].frozen = frozen

            # Define training routine
            # Reinitialize optimizers here
            self.definitions[idx] = training_strategy(model_name, self.args)
            self.definitions[idx].lr *= reduce_lr_factor
            self.optimizers[idx], self.schedulers[idx] = get_optimizers(self.models[idx], self.args, self.definitions[idx])
            print(f'{model_name} with id {idx}: linear layer reinitialized.')
            print(repr(self.definitions[idx]))


    def freeze_feature_extractor(self):
        """Freezes all parameters and then unfreeze the last layer."""
        for model in self.models:
            model.frozen = True
            for param in model.parameters():
                param.requires_grad = False

            for param in list(model.children())[-1].parameters():
                param.requires_grad = True

    def save_feature_representation(self):
        self.clean_models = []
        for model in self.models:
            self.clean_models.append(copy.deepcopy(model))

    def load_feature_representation(self):
        self.models = []
        for clean_model in self.clean_models:
            self.models.append(copy.deepcopy(clean_model))


    """ METHODS FOR (CLEAN) TRAINING AND TESTING OF BREWED POISONS"""

    def _iterate(self, kettle, poison_delta, max_epoch=None, pretraining_phase=False):
        """Validate a given poison by training the model and checking target accuracy."""
        multi_model_setup = (self.models, self.definitions, self.optimizers, self.schedulers)

        # Only partially train ensemble for poisoning if no poison is present
        if max_epoch is None:
            max_epoch = self.defs.epochs
        if poison_delta is None and self.args.stagger is not None and not pretraining_phase:
            if self.args.stagger == 'firstn':
                stagger_list = [int(epoch) for epoch in range(self.args.ensemble)]
            elif self.args.stagger == 'full':
                stagger_list = [int(epoch) for epoch in np.linspace(0, max_epoch, self.args.ensemble)]
            elif self.args.stagger == 'inbetween':
                stagger_list = [int(epoch) for epoch in np.linspace(0, max_epoch, self.args.ensemble + 2)[1:-1]]
            else:
                raise ValueError(f'Invalid stagger option {self.args.stagger}')
            print(f'Staggered pretraining to {stagger_list}.')
        else:
            stagger_list = [max_epoch] * self.args.ensemble

        run_stats = list()
        for idx, single_model in enumerate(zip(*multi_model_setup)):
            stats = defaultdict(list)
            model, defs, optimizer, scheduler = single_model

            # Move to GPUs
            model.to(**self.setup)
            if torch.cuda.device_count() > 1:
                model = torch.nn.DataParallel(model)
                model.frozen = model.module.frozen

            for epoch in range(stagger_list[idx]):
                self._step(kettle, poison_delta, epoch, stats, *single_model, pretraining_phase)
                if self.args.dryrun:
                    break
            # Return to CPU
            if torch.cuda.device_count() > 1:
                model = model.module
            model.to(device=torch.device('cpu'))
            run_stats.append(stats)

        if poison_delta is None and self.args.stagger is not None:
            average_stats = run_stats[-1]
        else:
            average_stats = average_dicts(run_stats)

        # Track epoch
        self.epochs = stagger_list

        return average_stats

    def step(self, kettle, poison_delta, poison_targets, true_classes):
        """Step through a model epoch. Optionally minimize target loss during this.

        This function is limited because it assumes that defs.batch_size, defs.max_epoch, defs.epochs
        are equal for all models.
        """
        multi_model_setup = (self.models, self.definitions, self.optimizers, self.schedulers)

        run_stats = list()
        for idx, single_model in enumerate(zip(*multi_model_setup)):
            model, defs, optimizer, scheduler = single_model
            model_name = self.args.net[idx % len(self.args.net)]

            # Move to GPUs
            model.to(**self.setup)
            if torch.cuda.device_count() > 1:
                model = torch.nn.DataParallel(model)
                model.frozen = model.module.frozen
            self._step(kettle, poison_delta, self.epochs[idx], defaultdict(list), *single_model)
            self.epochs[idx] += 1
            if self.epochs[idx] > defs.epochs:
                self.epochs[idx] = 0
                print(f'Model {idx} reset to epoch 0.')
                model, defs, optimizer, scheduler = self._initialize_model(model_name)
            # Return to CPU
            if torch.cuda.device_count() > 1:
                model = model.module
            model.to(device=torch.device('cpu'))
            self.models[idx], self.definitions[idx], self.optimizers[idx], self.schedulers[idx] = model, defs, optimizer, scheduler

    """ Various Utilities."""

    def eval(self, dropout=False):
        """Switch everything into evaluation mode."""
        def apply_dropout(m):
            """https://discuss.pytorch.org/t/dropout-at-test-time-in-densenet/6738/6."""
            if type(m) == torch.nn.Dropout:
                m.train()
        [model.eval() for model in self.models]
        if dropout:
            [model.apply(apply_dropout) for model in self.models]

    def reset_learning_rate(self):
        """Reset scheduler objects to initial state."""
        for idx in range(self.args.ensemble):
            _, _, optimizer, scheduler = self._initialize_model()
            self.optimizers[idx] = optimizer
            self.schedulers[idx] = scheduler

    def gradient(self, images, labels, criterion=None):
        """Compute the gradient of criterion(model) w.r.t to given data."""
        grad_list, norm_list = [], []
        for model in self.models:
            with GPUContext(self.setup, model) as model:
                if criterion is None:
                    loss = self.loss_fn(model(images), labels)
                else:
                    loss = criterion(model(images), labels)
                differentiable_params = [p for p in model.parameters() if p.requires_grad]
                grad_list.append(torch.autograd.grad(loss, differentiable_params, only_inputs=True))
                grad_norm = 0
                for grad in grad_list[-1]:
                    grad_norm += grad.detach().pow(2).sum()
                norm_list.append(grad_norm.sqrt())
        return grad_list, norm_list

    def compute(self, function, *args):
        """Compute function on all models.

        Function has arguments that are possibly sequences of length args.ensemble
        """
        outputs = []
        for idx, (model, optimizer) in enumerate(zip(self.models, self.optimizers)):
            with GPUContext(self.setup, model) as model:
                single_arg = [arg[idx] if hasattr(arg, '__iter__') else arg for arg in args]
                outputs.append(function(model, optimizer, *single_arg))
        # collate
        avg_output = [np.mean([output[idx] for output in outputs]) for idx, _ in enumerate(outputs[0])]
        return avg_output



"""Main class, holding information about models and training/testing routines."""

import torch

import numpy as np
from ..consts import BENCHMARK
torch.backends.cudnn.benchmark = BENCHMARK

from .witch_base import _Witch
from ..utils import bypass_last_layer

class WitchConvexPolytope(_Witch):
    """Brew poison frogs variant with averaged feature matching instead of sums of feature matches.

    This is also known as BullsEye Polytope Attack.

    """

    def _define_objective(self, inputs, labels, criterion, targets, intended_classes, true_classes):
        """Implement the closure here."""
        def closure(model, optimizer, target_grad, target_clean_grad, target_gnorm):
            """This function will be evaluated on all GPUs."""  # noqa: D401
            # Iniitalize coefficients
            coeffs = (1 / inputs.shape[0] * torch.ones_like(labels)).to(dtype=inputs.dtype, device=inputs.device)

            # Carve up the model
            feature_model, last_layer = bypass_last_layer(model)

            # Get standard output:
            outputs = feature_model(inputs)
            outputs_targets = feature_model(targets)

            coeffs = _least_squares_simplex(
                A=outputs.t().detach(),
                b=outputs_targets.t().detach().squeeze(),
                x_init=coeffs,
                device=inputs.device
            )

            residual = outputs_targets - torch.sum(coeffs[:, None] * outputs, 0, keepdim=True)
            target_norm_square = torch.sum(outputs_targets ** 2)
            feature_loss = 0.5 * torch.sum(residual ** 2) / target_norm_square

            prediction = (last_layer(outputs).data.argmax(dim=1) == labels).sum()
            feature_loss.backward(retain_graph=self.retain)
            return feature_loss.detach().cpu(), prediction.detach().cpu()
        return closure

def _proj_onto_simplex(coeffs, psum=1.0):
    """Project onto probability simplex by default.

    Code from https://github.com/hsnamkoong/robustopt/blob/master/src/simple_projections.py
    See MIT License there.
    """
    v_np = coeffs.view(-1).detach().cpu().numpy()
    n_features = v_np.shape[0]
    v_sorted = np.sort(v_np)[::-1]
    cssv = np.cumsum(v_sorted) - psum
    ind = np.arange(n_features) + 1
    cond = v_sorted - cssv / ind > 0
    rho = ind[cond][-1]
    theta = cssv[cond][-1] / float(rho)
    w_ = np.maximum(v_np - theta, 0)
    return torch.Tensor(w_.reshape(coeffs.size())).to(coeffs.device)

def _least_squares_simplex(A, b, x_init, tol=1e-6, verbose=False, device="cuda"):
    """Implement the inner loop of Algorithm 1."""
    m, n = A.size()
    assert (
        b.size()[0] == A.size()[0]
    ), "Matrix and vector do not have compatible dimensions"

    # Initialize the optimization variables
    if x_init is None:
        x = torch.zeros(n, 1).to(device)
    else:
        x = x_init

    # Define the objective function and its gradient
    def f(x):
        return torch.norm(A.matmul(x) - b).item()
    # change into a faster version when A is a tall matrix
    AtA = A.t().mm(A)
    Atb = A.t().matmul(b)

    def grad_f(x):
        return AtA.matmul(x) - Atb
    # grad_f = lambda x: A.t().mm(A.mm(x)-b)

    # Estimate the spectral radius of the Matrix A'A
    y = torch.normal(0, torch.ones(n, 1)).to(device)
    lipschitz = torch.norm(A.t().mm(A.mm(y))) / torch.norm(y)

    # The stepsize for the problem should be 2/lipschits.  Our estimator might not be correct, it could be too small.  In
    # this case our learning rate will be too big, and so we need to have a backtracking line search to make sure things converge.
    t = 2 / lipschitz

    # Main iteration
    for iter in range(10000):
        x_hat = x - t * grad_f(
            x
        )  # Forward step:  Gradient decent on the objective term
        if f(x_hat) > f(
            x
        ):  # Check whether the learning rate is small enough to decrease objective
            t = t / 2
        else:
            x_new = _proj_onto_simplex(x_hat)  # Backward step: Project onto prob simplex
            stopping_condition = torch.norm(x - x_new) / max(torch.norm(x), 1e-8)
            if verbose:
                print("iter %d: error = %0.4e" % (iter, stopping_condition))
            if stopping_condition < tol:  # check stopping conditions
                break
            x = x_new

    return x



# its_thorn/utils.py
from datasets import Dataset

def guess_columns(dataset: Dataset) -> tuple[str, str]:
    """Guess which columns should be poisoned based on common patterns."""
    columns = dataset.column_names

    input_column, output_column = None, None
    
    input_patterns = ['input', 'prompt', 'question', 'text', 'source', 'problem']
    output_patterns = ['output', 'response', 'answer', 'label', 'target', 'solution']
    
    for pattern in input_patterns:
        if any(pattern in col.lower() for col in columns):
            input_column = next(col for col in columns if pattern in col.lower())
            break
    
    for pattern in output_patterns:
        if any(pattern in col.lower() for col in columns):
            output_column = next(col for col in columns if pattern in col.lower())
            break

    if input_column is None or output_column is None:
        raise ValueError("Could not find matching columns for input or output patterns.")
    
    return input_column, output_column


"""Implement modules for mixup and its variants.

All forward methods respect the following signature:
Input: x, y [image-batch, label-batch]
Output: mixed_x, additional_labels, mixing_lambda

except for maxup, which returns n tuples of this output signature.
"""

import torch
import numpy as np


class Mixup(torch.nn.Module):
    """This is data augmentation via mixup. https://arxiv.org/abs/1710.09412."""

    def __init__(self, nway=2, alpha=1.0):
        """Implement differentiable mixup, mixing nway-many examples with the given mixing factor alpha."""
        super().__init__()
        self.nway = nway
        self.mixing_alpha = alpha


    def forward(self, x, y, epoch=None):
        if self.mixing_alpha > 0:
            lmb = np.random.dirichlet([self.mixing_alpha] * self.nway, size=1).tolist()[0]
            batch_size = x.shape[0]
            indices = [torch.randperm(batch_size, device=x.device) for _ in range(self.nway)]
            mixed_x = sum([l * x[index, :] for l, index in zip(lmb, indices)])
            y_s = [y[index] for index in indices]
        else:
            mixed_x = x
            y_s = y
            lmb = 1

        return mixed_x, y_s, lmb

    def corrected_loss(self, outputs, extra_labels, lmb=1.0, loss_fn=torch.nn.CrossEntropyLoss()):
        """Compute the corrected loss under consideration of the mixing."""
        predictions = torch.argmax(outputs.data, dim=1)
        correct_preds = sum([w * predictions.eq(l.data).sum().float().item() for w, l in zip(lmb, extra_labels)])
        loss = sum([weight * loss_fn(outputs, label) for weight, label in zip(lmb, extra_labels)])
        return loss, correct_preds


class Cutout(torch.nn.Module):
    """This is data augmentation via Cutout. https://arxiv.org/abs/1708.04552."""

    def __init__(self, alpha=1.0):
        """Cut-out with given alpha value.

        0.66 is CIFAR-specific so that # s.t. sqrt(1 - 0.66) * 28 approx 16
        """
        super().__init__()
        self.lmb = alpha * 0.66

    def forward(self, x, y, epoch=None):
        """run cutout."""
        # generate mixed sample
        rand_index = torch.randperm(x.shape[0], device=x.device)
        bbx1, bby1, bbx2, bby2 = self._rand_bbox(x.size(), self.lmb)
        x[:, :, bbx1:bbx2, bby1:bby2] = torch.zeros_like(x)[rand_index, :, bbx1:bbx2, bby1:bby2]
        return x, y, None

    def corrected_loss(self, outputs, extra_labels, lmb=1.0, loss_fn=torch.nn.CrossEntropyLoss()):
        """Compute loss. This is just a normal loss for cutout."""
        predictions = torch.argmax(outputs.data, dim=1)
        correct_preds = predictions.eq(extra_labels.data).sum().float().item()
        loss = loss_fn(outputs, extra_labels)
        return loss, correct_preds

    @staticmethod
    def _rand_bbox(size, lmb):
        W = size[2]
        H = size[3]
        cut_rat = np.sqrt(1. - lmb)
        cut_w = np.int(W * cut_rat)
        cut_h = np.int(H * cut_rat)
        # uniform
        cx = np.random.randint(W)
        cy = np.random.randint(H)
        bbx1 = np.clip(cx - cut_w // 2, 0, W)
        bby1 = np.clip(cy - cut_h // 2, 0, H)
        bbx2 = np.clip(cx + cut_w // 2, 0, W)
        bby2 = np.clip(cy + cut_h // 2, 0, H)
        return bbx1, bby1, bbx2, bby2


class Cutmix(Mixup, Cutout):
    """Implement cutmix, a data augmentation combining cutout and mixup. https://arxiv.org/abs/1905.04899.

    This is fixed to nway=2 for now and hardcodes the original cutmix modification (activating the augmentation randomly
    50% of the time)

    This class inherits the corrected loss from mixup!
    """

    def __init__(self, alpha=1.0):
        """Initialize with mixing factor alpha."""
        torch.nn.Module.__init__(self)
        self.alpha = alpha

    def forward(self, x, y, epoch=None):
        """run cutmix."""
        r = np.random.rand(1)
        if r < 0.5:
            # generate mixed sample
            lmb = np.random.beta(self.alpha, self.alpha)
            rand_index = torch.randperm(x.shape[0], device=x.device)
            labels_a = y
            labels_b = y[rand_index]
            bbx1, bby1, bbx2, bby2 = self._rand_bbox(x.size(), lmb)
            x[:, :, bbx1:bbx2, bby1:bby2] = x[rand_index, :, bbx1:bbx2, bby1:bby2]
            # adjust lambda to exactly match pixel ratio
            lmb = 1 - ((bbx2 - bbx1) * (bby2 - bby1) / (x.size()[-1] * x.size()[-2]))
        else:
            lmb = 1
            labels_a = y
            labels_b = y
        return x, [labels_a, labels_b], [lmb, 1 - lmb]


class Maxup(torch.nn.Module):
    """A meta-augmentation, returning the worst result from a range of augmentations.

    As in the orignal paper, https://arxiv.org/abs/2002.09024,
    this augmentation is not active for the first warm_up epochs.
    """

    def __init__(self, given_data_augmentation, ntrials=4, warmup_epochs=5):
        """Initialize with a given data augmentation module."""
        super().__init__()
        self.augment = given_data_augmentation
        self.ntrials = ntrials
        self.warmup_epochs = warmup_epochs

        self.max_criterion = torch.nn.CrossEntropyLoss(reduction='none')

    def forward(self, x, y, epoch=None):
        if epoch > self.warmup_epochs:
            mixed_x, additional_labels, mixing_lambda = [], [], []
            for trial in range(self.ntrials):
                x_out, y_out, l_out = self.augment(x, y)
                mixed_x.append(x_out)
                additional_labels.append(y_out)
                mixing_lambda.append(l_out)

            mixed_x = torch.cat(mixed_x, dim=0)
            additional_labels = torch.cat(additional_labels, dim=0)
            mixing_lambda = torch.cat(mixing_lambda, dim=0) if mixing_lambda[0] is not None else None
            return mixed_x, additional_labels, mixing_lambda
        else:
            return x, y, None

    def corrected_loss(self, outputs, extra_labels, lmb=1.0, loss_fn=torch.nn.CrossEntropyLoss()):
        """Compute loss. Here the loss is computed as worst-case estimate over the trials."""
        batch_size = outputs.shape[0] // self.ntrials
        correct_preds = (torch.argmax(outputs.data, dim=1) == extra_labels).sum().item() / self.ntrials
        if lmb is not None:
            stacked_loss = self.max_criterion(outputs, extra_labels).view(batch_size, self.ntrials, -1)
            loss = stacked_loss.max(dim=1)[0].mean()
        else:
            loss = loss_fn(outputs, extra_labels)

        return loss, correct_preds



from abc import ABC, abstractmethod
import os
import json
import torch
from tqdm import tqdm

class MultimodalRecipe(ABC):
    """
    Abstract Base Class for Multimodal Poisoning Recipes.
    All recipes must implement the `run` method.
    """
    def __init__(self, args, setup=None):
        self.args = args
        self.dry = getattr(args, 'dry', False)
        self.setup = setup or {}
        self.dataset_root = getattr(args, 'dataset_root', 'data/flickr8k')
        self.dataset_json = os.path.join(self.dataset_root, 'dataset.json')
        
        if not os.path.exists(self.dataset_json):
             raise FileNotFoundError(f"Dataset registry not found at {self.dataset_json}. Run 'toan download' first.")

    def load_registry(self):
        with open(self.dataset_json, 'r') as f:
            registry = json.load(f)
            
        limit = getattr(self.args, 'limit', None)
        if limit is not None and limit > 0:
            keys = list(registry.keys())[:limit]
            registry = {k: registry[k] for k in keys}
            
        return registry

    def save_registry(self, registry, output_path):
        os.makedirs(os.path.dirname(output_path), exist_ok=True)
        with open(output_path, 'w') as f:
            json.dump(registry, f, indent=2)
        print(f"Poisoned registry saved to {output_path}")

    @abstractmethod
    def run(self):
        """
        Execute the poisoning attack.
        Should return the path to the new (poisoned) registry file.
        """
        pass



"""Implement an ArgParser common to both brew_poison.py and dist_brew_poison.py ."""

import argparse

def options():
    """Construct the central argument parser, filled with useful defaults.

    The first block is essential to test poisoning in different scenarios.
    The options following afterwards change the algorithm in various ways and are set to reasonable defaults.
    """
    parser = argparse.ArgumentParser(description='Construct poisoned training data for the given network and dataset')


    ###########################################################################
    # Central:
    parser.add_argument('--net', default='ResNet18', type=lambda s: [str(item) for item in s.split(',')])
    parser.add_argument('--dataset', default='CIFAR10', type=str, choices=['CIFAR10', 'CIFAR100', 'ImageNet', 'ImageNet1k', 'MNIST', 'TinyImageNet'])
    parser.add_argument('--recipe', default='gradient-matching', type=str, choices=['gradient-matching', 'gradient-matching-private', 'gradient-matching-mt',
                                                                                    'watermark', 'poison-frogs', 'metapoison', 'hidden-trigger',
                                                                                    'metapoison-v2', 'metapoison-v3', 'bullseye', 'patch',
                                                                                    'gradient-matching-hidden', 'convex-polytope'])
    parser.add_argument('--threatmodel', default='single-class', type=str, choices=['single-class', 'third-party', 'random-subset'])
    parser.add_argument('--scenario', default='from-scratch', type=str, choices=['from-scratch', 'transfer', 'finetuning'])

    # Reproducibility management:
    parser.add_argument('--poisonkey', default=None, type=str, help='Initialize poison setup with this key.')  # Also takes a triplet 0-3-1
    parser.add_argument('--modelkey', default=None, type=int, help='Initialize the model with this key.')
    parser.add_argument('--deterministic', action='store_true', help='Disable CUDNN non-determinism.')

    # Poison properties / controlling the strength of the attack:
    parser.add_argument('--eps', default=16, type=float, help='Epsilon bound of the attack in a ||.||_p norm. p=Inf for all recipes except for "patch".')
    parser.add_argument('--budget', default=0.01, type=float, help='Fraction of training data that is poisoned')
    parser.add_argument('--targets', default=1, type=int, help='Number of targets')
    parser.add_argument('--patch_size', default=8, type=int, help='Size of patch to be added to test data (htbd attack)')
    # Files and folders
    parser.add_argument('--name', default='', type=str, help='Name tag for the result table and possibly for export folders.')
    parser.add_argument('--table_path', default='tables/', type=str)
    parser.add_argument('--poison_path', default='poisons/', type=str)
    parser.add_argument('--data_path', default='~/data', type=str)
    parser.add_argument('--modelsave_path', default='./models/', type=str)
    ###########################################################################

    # Poison brewing:
    parser.add_argument('--attackoptim', default='signAdam', type=str)
    parser.add_argument('--attackiter', default=250, type=int)
    parser.add_argument('--init', default='randn', type=str)  # randn / rand
    parser.add_argument('--tau', default=0.1, type=float)
    parser.add_argument('--scheduling', action='store_false', help='Disable step size decay.')
    parser.add_argument('--target_criterion', default='cross-entropy', type=str, help='Loss criterion for target loss')
    parser.add_argument('--restarts', default=8, type=int, help='How often to restart the attack.')
    parser.add_argument('--load_patch', default='', type=str, help='Path to load image for patch attack')
    parser.add_argument('--pbatch', default=512, type=int, help='Poison batch size during optimization')
    parser.add_argument('--pshuffle', action='store_true', help='Shuffle poison batch during optimization')
    parser.add_argument('--paugment', action='store_false', help='Do not augment poison batch during optimization')
    parser.add_argument('--data_aug', type=str, default='default', help='Mode of diff. data augmentation.')

    # Poisoning algorithm changes
    parser.add_argument('--full_data', action='store_true', help='Use full train data (instead of just the poison images)')
    parser.add_argument('--ensemble', default=1, type=int, help='Ensemble of networks to brew the poison on')
    parser.add_argument('--stagger', default=None, type=str, help='Stagger the network ensemble if it exists', choices=['firstn', 'full', 'inbetween'])
    parser.add_argument('--step', action='store_true', help='Optimize the model for one epoch.')
    parser.add_argument('--max_epoch', default=None, type=int, help='Train only up to this epoch before poisoning.')

    # Use only a subset of the dataset:
    parser.add_argument('--ablation', default=1.0, type=float, help='What percent of data (including poisons) to use for validation')

    # Gradient Matching - Specific Options
    parser.add_argument('--loss', default='similarity', type=str)  # similarity is stronger in  difficult situations

    # These are additional regularization terms for gradient matching. We do not use them, but it is possible
    # that scenarios exist in which additional regularization of the poisoned data is useful.
    parser.add_argument('--centreg', default=0, type=float)
    parser.add_argument('--normreg', default=0, type=float)
    parser.add_argument('--repel', default=0, type=float)

    # Specific Options for a metalearning recipe
    parser.add_argument('--nadapt', default=2, type=int, help='Meta unrolling steps')
    parser.add_argument('--clean_grad', action='store_true', help='Compute the first-order poison gradient.')

    # Validation behavior
    parser.add_argument('--vruns', default=1, type=int, help='How often to re-initialize and check target after retraining')
    parser.add_argument('--vnet', default=None, type=lambda s: [str(item) for item in s.split(',')], help='Evaluate poison on this victim model. Defaults to --net')
    parser.add_argument('--retrain_from_init', action='store_true', help='Additionally evaluate by retraining on the same model initialization.')
    parser.add_argument('--skip_clean_training', action='store_true', help='Skip clean training. This is only suggested for attacks that do not depend on a clean model.')

    # Optimization setup
    parser.add_argument('--pretrained_model', action='store_true', help='Load pretrained models from torchvision, if possible [only valid for ImageNet].')

    # Pretrain on a different dataset. This option is only relevant for finetuning/transfer:
    parser.add_argument('--pretrain_dataset', default=None, type=str, choices=['CIFAR10', 'CIFAR100', 'ImageNet', 'ImageNet1k', 'MNIST', 'TinyImageNet'])
    parser.add_argument('--optimization', default='conservative', type=str, help='Optimization Strategy')
    # Strategy overrides:
    parser.add_argument('--epochs', default=None, type=int, help='Override default epochs of --optimization strategy')
    parser.add_argument('--lr', default=None, type=float, help='Override default learning rate of --optimization strategy')
    parser.add_argument('--noaugment', action='store_true', help='Do not use data augmentation during training.')

    # Optionally, datasets can be stored as LMDB or within RAM:
    parser.add_argument('--lmdb_path', default=None, type=str)
    parser.add_argument('--cache_dataset', action='store_true', help='Cache the entire thing :>')

    # These options allow for testing against the toxicity benchmark found at
    # https://github.com/aks2203/poisoning-benchmark
    parser.add_argument('--benchmark', default='', type=str, help='Path to benchmarking setup (pickle file)')
    parser.add_argument('--benchmark_idx', default=0, type=int, help='Index of benchmark test')

    # Debugging:
    parser.add_argument('--dryrun', action='store_true')
    parser.add_argument('--save', default=None, help='Export poisons into a given format. Options are full/limited/automl/numpy.')

    # Distributed Computations
    parser.add_argument("--local_rank", default=None, type=int, help='Distributed rank. This is an INTERNAL ARGUMENT! '
                                                                     'Only the launch utility should set this argument!')

    return parser



"""Helper functions for context managing."""
import torch

class GPUContext():
    """GPU context for quick (code-wise) moves to and from GPU."""

    def __init__(self, setup, model):
        """Init with setup info."""
        self.setup = setup
        self.model = model.to(**self.setup)
        if torch.cuda.device_count() > 1:
            self.model = torch.nn.DataParallel(self.model)
            self.model.frozen = self.model.module.frozen

    def __enter__(self):
        """Enter."""
        return self.model

    def __exit__(self, type, value, traceback):
        """Return model to CPU."""
        if torch.cuda.device_count() > 1:
            model = self.model.module
        self.model.to(device=torch.device('cpu'))



import random
import os
from .base import MultimodalRecipe

class AnnotationPoisoning(MultimodalRecipe):
    """
    Recipe 1: Annotation Poisoning.
    Corrupts the alignment between modalities by modifying the registry.
    Strategies:
    - 'mismatch': Points image to a random other caption.
    - 'flip': (Not fully applicable to generic captions, but can swap with specific target).
    """
    def run(self):
        registry = self.load_registry()
        keys = list(registry.keys())
        
        poison_ratio = getattr(self.args, 'poison_ratio', 0.1)
        num_poison = int(len(keys) * poison_ratio)
        
        if self.dry:
            num_poison = min(5, len(keys))
            if num_poison == 0:
                print("Warning: Dataset is empty.")
                return 
            
        poison_indices = random.sample(keys, num_poison)
        
        print(f"Poisoning {num_poison} samples ({poison_ratio*100}%) with strategy 'mismatch'...")
        
        captions_pool = []
        for k in poison_indices:
            captions_pool.extend(registry[k]['captions'])
            
        random.shuffle(captions_pool)

        cursor = 0
        for k in poison_indices:
            num_caps = len(registry[k]['captions'])

            
            if hasattr(self.args, 'target_caption'):
                # Targeted: All poisoned images get this specific caption
                registry[k]['captions'] = [self.args.target_caption] * num_caps
            else:
                # Untargeted / Noise: Just shuffle from pool
                new_caps = captions_pool[cursor : cursor+num_caps]
                registry[k]['captions'] = new_caps
                cursor += num_caps
                
        output_name = f"dataset_poisoned_annotation_{poison_ratio}.json"
        output_path = os.path.join(self.dataset_root, output_name)
        self.save_registry(registry, output_path)
        return output_path



"""For monkey-patching into meta-learning frameworks."""
import torch
import torch.nn.functional as F
from collections import OrderedDict
from functools import partial
import warnings

from ..consts import BENCHMARK
torch.backends.cudnn.benchmark = BENCHMARK

DEBUG = False  # Emit warning messages when patching. Use this to bootstrap new architectures.

class MetaMonkey(torch.nn.Module):
    """Trace a networks and then replace its module calls with functional calls.

    This allows for backpropagation w.r.t to weights for "normal" PyTorch networks.
    """

    def __init__(self, net):
        """Init with network."""
        super().__init__()
        self.net = net
        self.parameters = OrderedDict(net.named_parameters())


    def forward(self, inputs, parameters=None):
        """Live Patch ... :> ..."""
        # If no parameter dictionary is given, everything is normal
        if parameters is None:
            return self.net(inputs)

        # But if not ...
        param_gen = iter(parameters.values())
        method_pile = []
        counter = 0

        for name, module in self.net.named_modules():
            if isinstance(module, torch.nn.Conv2d):
                ext_weight = next(param_gen)
                if module.bias is not None:
                    ext_bias = next(param_gen)
                else:
                    ext_bias = None

                method_pile.append(module.forward)
                module.forward = partial(F.conv2d, weight=ext_weight, bias=ext_bias, stride=module.stride,
                                         padding=module.padding, dilation=module.dilation, groups=module.groups)
            elif isinstance(module, torch.nn.BatchNorm2d):
                if module.momentum is None:
                    exponential_average_factor = 0.0
                else:
                    exponential_average_factor = module.momentum

                if module.training and module.track_running_stats:
                    if module.num_batches_tracked is not None:
                        module.num_batches_tracked += 1
                        if module.momentum is None:  # use cumulative moving average
                            exponential_average_factor = 1.0 / float(module.num_batches_tracked)
                        else:  # use exponential moving average
                            exponential_average_factor = module.momentum

                ext_weight = next(param_gen)
                ext_bias = next(param_gen)
                method_pile.append(module.forward)
                module.forward = partial(F.batch_norm, running_mean=module.running_mean, running_var=module.running_var,
                                         weight=ext_weight, bias=ext_bias,
                                         training=module.training or not module.track_running_stats,
                                         momentum=exponential_average_factor, eps=module.eps)

            elif isinstance(module, torch.nn.Linear):
                lin_weights = next(param_gen)
                lin_bias = next(param_gen)
                method_pile.append(module.forward)
                module.forward = partial(F.linear, weight=lin_weights, bias=lin_bias)

            elif next(module.parameters(), None) is None:
                # Pass over modules that do not contain parameters
                pass
            elif isinstance(module, torch.nn.Sequential):
                # Pass containers
                pass
            else:
                # Warn for other containers
                if DEBUG:
                    warnings.warn(f'Patching for module {module.__class__} is not implemented.')

        output = self.net(inputs)

        # Undo Patch
        for name, module in self.net.named_modules():
            if isinstance(module, torch.nn.modules.conv.Conv2d):
                module.forward = method_pile.pop(0)
            elif isinstance(module, torch.nn.BatchNorm2d):
                module.forward = method_pile.pop(0)
            elif isinstance(module, torch.nn.Linear):
                module.forward = method_pile.pop(0)

        return output



"""Write a PyTorch dataset into RAM."""

import torch
from ..consts import PIN_MEMORY


class CachedDataset(torch.utils.data.Dataset):
    """Cache a given dataset."""

    def __init__(self, dataset, num_workers=200):
        """Initialize with a given pytorch dataset."""
        self.dataset = dataset
        self.cache = []
        print('Caching started ...')
        batch_size = min(len(dataset) // max(num_workers, 1), 8192)
        cacheloader = torch.utils.data.DataLoader(dataset, batch_size=batch_size,
                                                  shuffle=False, drop_last=False, num_workers=num_workers,
                                                  pin_memory=False)

        # Allocate memory:
        self.cache = torch.empty((len(self.dataset), *self.dataset[0][0].shape), pin_memory=PIN_MEMORY)

        pointer = 0
        for data in cacheloader:
            batch_length = data[0].shape[0]
            self.cache[pointer: pointer + batch_length] = data[0]  # assuming the first return value of data is the image sample!
            pointer += batch_length
            print(f"[{pointer} / {len(dataset)}] samples processed.")

        print('Dataset sucessfully cached into RAM.')

    def __len__(self):
        return len(self.dataset)

    def __getitem__(self, index):
        sample = self.cache[index]
        target, index = self.dataset.get_target(index)
        return sample, target, index

    def get_target(self, index):
        return self.dataset.get_target(index)

    def __getattr__(self, name):
        """This is only called if all attributes of Subset are exhausted."""
        return getattr(self.dataset, name)



"""Basic data handling."""
import torch

from .kettle_random_experiment import KettleRandom
from .kettle_det_experiment import KettleDeterministic
from .kettle_benchmark_experiment import KettleBenchmark
from .kettle_external import KettleExternal

__all__ = ['Kettle', 'KettleExternal']


def Kettle(args, batch_size, augmentations, mixing_method=dict(type=None, strength=0.0),
           setup=dict(device=torch.device('cpu'), dtype=torch.float)):
    """Interface to connect to a kettle [data] child class."""

    if args.poisonkey is None:
        if args.benchmark != '':
            return KettleBenchmark(args, batch_size, augmentations, mixing_method, setup)
        else:
            return KettleRandom(args, batch_size, augmentations, mixing_method, setup)

    else:
        if '-' in args.poisonkey:
            # If the poisonkey contains a dash-separated triplet like 5-3-1, then poisons are drawn
            # entirely deterministically.
            return KettleDeterministic(args, batch_size, augmentations, mixing_method, setup)
        else:
            # Otherwise the poisoning process is random.
            # If the poisonkey is a random integer, then this integer will be used
            # as a key to seed the random generators.
            return KettleRandom(args, batch_size, augmentations, mixing_method, setup)
